{-# OPTIONS_GHC 
                -XMultiParamTypeClasses 
                -XFlexibleInstances
                -XUndecidableInstances
                -XOverlappingInstances
                -XFlexibleContexts
#-}

module AutoLift (
  MonadModel(..),MonadAlgModel(..),MonadExtModel(..),
  get, put, ask, inEnv, abort,callCC,throw,handle,trace,zeroND,plusND, step,
  module Monatron
) where

import Monatron
import Control.Monad
import Data.Monoid
  
----------------------------
-- Automatic liftings
----------------------------
  
class (Functor f, Monad m) => MonadModel f m where
    model :: Model f m


class (Functor f, Monad m) => MonadAlgModel f m where
    algModel :: AlgModel f m

class (Functor f, Functor g, Monad m) => MonadExtModel f g m where
    extModel :: ExtModel f g m

instance  (FMonadT t, MonadModel f m, Monad (t m), Functor m, Monad (t (Cod m)))
          =>  MonadModel f (t m) where
    model = liftModel model

instance  (MonadT t, MonadAlgModel f m, Monad (t m))
          => MonadAlgModel f (t m) where
    algModel = liftAlgModel algModel

instance  (MMonadT t, MonadExtModel f g m, Monad (t m), Functor m, 
          Functor (t f), Functor (t m)) => MonadExtModel f g (t m) where
    extModel = liftExtModel extModel

ask  :: MonadModel (ReaderOp s) m => m s
ask  = askX model

inEnv  :: MonadModel (ReaderOp s) m => s -> m a -> m a
inEnv  = inEnvX model

instance Monad m => MonadModel (ReaderOp s) (ReaderT s m) where
    model = modelReaderT
  
get  :: MonadAlgModel (StateOp s) m => m s
get  = getX algModel

put  :: MonadAlgModel (StateOp s) m => s -> m ()
put  = putX algModel

instance Monad m => MonadAlgModel (StateOp s) (StateT s m) where
    algModel = modelStateT
  
abort  :: MonadAlgModel (ContOp r) m => r -> m a
abort  = abortX algModel

callCC  :: MonadAlgModel (ContOp r) m => ((a -> r) -> a) -> m a
callCC  = callCCX algModel

instance Monad m => MonadAlgModel (ContOp (m r)) (ContT r m) where
    algModel = modelContT
  
throw  :: MonadAlgModel (ThrowOp x) m => x -> m a
throw  = throwX algModel

handle  :: MonadModel (HandleOp x) m => m a -> (x -> m a) -> m a
handle  = handleX model

instance Monad m => MonadModel (HandleOp x) (ExcT x m) where
    model = modelHandleExcT

instance Monad m => MonadAlgModel (ThrowOp x) (ExcT x m) where
    algModel = modelThrowExcT
  
trace  :: MonadAlgModel (WriterOp w) m => w -> m ()
trace  = traceX algModel

instance (Monoid w, Monad m) => MonadAlgModel (WriterOp w) (WriterT w m) where
    algModel = modelWriterT
  
step  :: (Functor f, MonadModel (StepOp f) m) => f (m x) -> m x
step  = stepX model

instance (Functor f, Monad m) => MonadModel (StepOp f) (StepT f m) where
   model = modelStepT

zeroND  :: MonadAlgModel NonDetOp m => m a
zeroND  = zeroNDX algModel

plusND  :: MonadAlgModel NonDetOp m => m a -> m a -> m a
plusND  = plusNDX algModel

instance Monad m => MonadAlgModel NonDetOp (NonDetT m) where
    algModel = modelNonDetT
  
